from pyaprilaire.const import Action, FunctionalDomain
from pyaprilaire.packet import NackPacket, Packet

import unittest


class Test_Packet(unittest.TestCase):
    def test_invalid_action(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 7, 1, 1, 0]))

        self.assertEqual(len(packets), 0)

    def test_invalid_functional_domain(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 3, 17, 1, 0]))

        self.assertEqual(len(packets), 0)

    def test_unmapped(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 3, 13, 1, 0]))

        self.assertEqual(len(packets), 0)

    def test_packet_empty_parse(self):
        packets = list(Packet.parse(bytes()))

        self.assertEqual(len(packets), 0)

    def test_packet_single_parse(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

        self.assertEqual(len(packets), 1)

    def test_packet_single_action(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

        packet = packets[0]

        self.assertEqual(packet.action, Action.READ_REQUEST)

    def test_packet_single_functional_domain(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

        packet = packets[0]

        self.assertEqual(packet.functional_domain, FunctionalDomain.SETUP)

    def test_packet_single_attribute(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 2, 1, 1, 107]))

        packet = packets[0]

        self.assertEqual(packet.attribute, 1)

    def test_packet_single_extra_data(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 8, 3, 2, 1, 1, 1, 1, 1, 10, 146])
        )

        self.assertEqual(len(packets), 1)

    def test_packet_multiple_parse(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
        )

        self.assertEqual(len(packets), 2)

    def test_packet_multiple_action(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
        )

        packet = packets[1]

        self.assertEqual(packet.action, Action.READ_RESPONSE)

    def test_packet_multiple_functional_domain(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
        )

        packet = packets[1]

        self.assertEqual(packet.functional_domain, FunctionalDomain.SCHEDULING)

    def test_packet_multiple_attribute(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 3, 2, 1, 1, 107, 1, 2, 0, 3, 3, 3, 4, 248])
        )

        packet = packets[1]

        self.assertEqual(packet.attribute, 4)

    def test_nack_parse(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 2, 6, 1, 0]))

        self.assertEqual(len(packets), 1)

    def test_nack_and_packet_parse(self):
        packets: list[Packet] = list(
            Packet.parse(
                [
                    0x01,
                    0x04,
                    0x00,
                    0x02,
                    0x06,
                    0x03,
                    0xCD,
                    0x01,
                    0x01,
                    0x00,
                    0x11,
                    0x03,
                    0x08,
                    0x02,
                    0xB4,
                    0x82,
                    0x55,
                    0x50,
                    0x93,
                    0x6D,
                    0x01,
                    0x49,
                    0x02,
                    0x01,
                    0x02,
                    0x0D,
                    0x04,
                    0x0E,
                    0x51,
                ]
            )
        )

        self.assertEqual(len(packets), 2)

    def test_nack_packet_parse(self):
        packets: list[Packet] = list(Packet.parse([1, 1, 0, 3, 6, 1, 0]))

        self.assertIsInstance(packets[0], NackPacket)

    def test_control_1_parse(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 7, 3, 2, 1, 1, 2, 10, 20, 107])
        )

        packet = packets[0]

        self.assertDictEqual(
            packet.data,
            {
                "mode": 1,
                "fan_mode": 2,
                "heat_setpoint": 10,
                "cool_setpoint": 20,
            },
        )

    def test_scheduling_4_parse(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 13, 3, 3, 4, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 42])
        )

        packet = packets[0]

        self.assertDictEqual(
            packet.data,
            {
                "hold": 1,
            },
        )

    def test_sensor_2_parse(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 11, 3, 5, 2, 1, 10, 2, 20, 3, 50, 4, 60, 12])
        )

        packet = packets[0]

        self.assertDictEqual(
            packet.data,
            {
                "indoor_temperature_controlling_sensor_status": 1,
                "indoor_temperature_controlling_sensor_value": 10,
                "outdoor_temperature_controlling_sensor_status": 2,
                "outdoor_temperature_controlling_sensor_value": 20,
                "indoor_humidity_controlling_sensor_status": 3,
                "indoor_humidity_controlling_sensor_value": 50,
                "outdoor_humidity_controlling_sensor_status": 4,
                "outdoor_humidity_controlling_sensor_value": 60,
            },
        )

    def test_identification_2_parse(self):
        packets: list[Packet] = list(
            Packet.parse([1, 1, 0, 9, 3, 8, 2, 1, 2, 3, 4, 5, 6, 176])
        )

        packet = packets[0]

        self.assertDictEqual(
            packet.data,
            {"mac_address": "1:2:3:4:5:6"},
        )

    def test_identification_4_parse(self):
        packets: list[Packet] = list(
            Packet.parse(
                [
                    1,
                    1,
                    0,
                    27,
                    3,
                    8,
                    4,
                    49,
                    50,
                    51,
                    52,
                    53,
                    0,
                    0,
                    0,
                    84,
                    101,
                    115,
                    116,
                    32,
                    78,
                    97,
                    109,
                    101,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    180,
                ]
            )
        )

        packet = packets[0]

        self.assertDictEqual(
            packet.data,
            {"location": "12345", "name": "Test Name"},
        )

    def test_decode_temperature(self):
        temperature = Packet._decode_temperature(0x15)

        self.assertEqual(temperature, 21)

    def test_decode_temperature_negative(self):
        temperature = Packet._decode_temperature(0x95)

        self.assertEqual(temperature, -21)

    def test_decode_temperature_decimal(self):
        temperature = Packet._decode_temperature(0x5A)

        self.assertEqual(temperature, 26.5)

    def test_decode_temperature_negative_decimal(self):
        temperature = Packet._decode_temperature(0xDA)

        self.assertEqual(temperature, -26.5)

    def test_decode_humidity_zero(self):
        humidity = Packet._decode_humidity(0)

        self.assertEqual(humidity, None)

    def test_decode_humidity_1(self):
        humidity = Packet._decode_humidity(1)

        self.assertEqual(humidity, 1)

    def test_decode_humidity_99(self):
        humidity = Packet._decode_humidity(99)

        self.assertEqual(humidity, 99)

    def test_decode_humidity_100(self):
        humidity = Packet._decode_humidity(100)

        self.assertEqual(humidity, None)

    def test_decode_humidity_nonzero(self):
        humidity = Packet._decode_humidity(50)

        self.assertEqual(humidity, 50)

    def test_decode_humidity_negative(self):
        humidity = Packet._decode_humidity(-50)

        self.assertEqual(humidity, None)

    def test_encode_temperature(self):
        encoded_temperature = Packet._encode_temperature(21)

        self.assertEqual(encoded_temperature, 0x15)

    def test_encode_temperature_fraction(self):
        encoded_temperature = Packet._encode_temperature(26.5)

        self.assertEqual(encoded_temperature, 0x5A)

    def test_encode_temperature_negative(self):
        encoded_temperature = Packet._encode_temperature(-21)

        self.assertEqual(encoded_temperature, 0x95)

    def test_encode_temperature_negative_fraction(self):
        encoded_temperature = Packet._encode_temperature(-26.5)

        self.assertEqual(encoded_temperature, 0xDA)

    def test_serialize_nack(self):
        serialized = NackPacket(2).serialize()

        self.assertSequenceEqual(
            serialized,
            [1, 0, 0, 2, 6, 2, 227],
        )

    def test_serialize_raw(self):
        serialized = Packet(
            Action.READ_REQUEST, FunctionalDomain.CONTROL, 1, 1, 1, raw_data=[1, 2, 3]
        ).serialize()

        self.assertSequenceEqual(
            serialized,
            [1, 1, 0, 6, 2, 2, 1, 1, 2, 3, 133],
        )

    def test_serialize_single_packet_no_data(self):
        serialized = Packet(
            Action.READ_REQUEST,
            FunctionalDomain.CONTROL,
            1,
            1,
            1,
        ).serialize()

        self.assertSequenceEqual(
            serialized,
            [1, 1, 0, 3, 2, 2, 1, 70],
        )

    def test_serialize_single_packet_no_data(self):
        serialized = Packet(
            Action.READ_REQUEST,
            FunctionalDomain.CONTROL,
            1,
            1,
            1,
        ).serialize()

        self.assertSequenceEqual(
            serialized,
            [1, 1, 0, 3, 2, 2, 1, 70],
        )

    def test_control_1_serialize(self):
        serialized = Packet(
            Action.READ_RESPONSE,
            FunctionalDomain.CONTROL,
            1,
            1,
            1,
            data={
                "mode": 1,
                "fan_mode": 2,
                "heat_setpoint": 10,
                "cool_setpoint": 20,
            },
        ).serialize()

        self.assertSequenceEqual(serialized, [1, 1, 0, 7, 3, 2, 1, 1, 2, 10, 20, 107])

    def test_scheduling_4_serialize(self):
        serialized = Packet(
            Action.READ_RESPONSE,
            FunctionalDomain.SCHEDULING,
            4,
            1,
            1,
            data={
                "hold": 1,
            },
        ).serialize()

        self.assertSequenceEqual(
            serialized, [1, 1, 0, 13, 3, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55]
        )

    def test_sensor_2_serialize(self):
        serialized = Packet(
            Action.READ_RESPONSE,
            FunctionalDomain.SENSORS,
            2,
            1,
            1,
            data={
                "indoor_temperature_controlling_sensor_status": 1,
                "indoor_temperature_controlling_sensor_value": 10,
                "outdoor_temperature_controlling_sensor_status": 2,
                "outdoor_temperature_controlling_sensor_value": 20,
                "indoor_humidity_controlling_sensor_status": 3,
                "indoor_humidity_controlling_sensor_value": 50,
                "outdoor_humidity_controlling_sensor_status": 4,
                "outdoor_humidity_controlling_sensor_value": 60,
            },
        ).serialize()

        self.assertSequenceEqual(
            serialized, [1, 1, 0, 11, 3, 5, 2, 1, 10, 2, 20, 3, 50, 4, 60, 12]
        )

    def test_identification_2_serialize(self):
        serialized = Packet(
            Action.READ_RESPONSE,
            FunctionalDomain.IDENTIFICATION,
            2,
            1,
            1,
            data={"mac_address": [1, 2, 3, 4, 5, 6]},
        ).serialize()

        self.assertSequenceEqual(
            serialized, [1, 1, 0, 9, 3, 8, 2, 1, 2, 3, 4, 5, 6, 176]
        )

    def test_identification_4_serialize(self):
        serialized = Packet(
            Action.READ_RESPONSE,
            FunctionalDomain.IDENTIFICATION,
            4,
            1,
            1,
            data={"location": "12345", "name": "Test Name"},
        ).serialize()

        self.assertSequenceEqual(
            serialized,
            [
                1,
                1,
                0,
                27,
                3,
                8,
                4,
                49,
                50,
                51,
                52,
                53,
                0,
                0,
                0,
                84,
                101,
                115,
                116,
                32,
                78,
                97,
                109,
                101,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                180,
            ],
        )



"""Client for interfacing with the thermostat"""

from __future__ import annotations

import asyncio

from collections.abc import Callable
from logging import Logger
from typing import Any

from .const import Action, FunctionalDomain, QUEUE_FREQUENCY
from .packet import NackPacket, Packet
from .socket_client import SocketClient


class _AprilaireClientProtocol(asyncio.Protocol):
    """Protocol for interacting with the thermostat over socket connection"""

    def __init__(
        self,
        data_received_callback: Callable[[FunctionalDomain, int, dict[str, Any]], None],
        reconnect_action: Callable[[], None],
        logger: Logger,
    ) -> None:
        """Initialize the protocol"""
        self.data_received_callback = data_received_callback
        self.reconnect_action = reconnect_action
        self.logger = logger

        self.transport: asyncio.Transport = None

        self.packet_queue = asyncio.Queue()

        self.sequence = 0

    def _get_sequence(self):
        self.sequence = (self.sequence + 1) % 128

        return self.sequence

    async def _send_packet(self, packet: Packet) -> None:
        """Send a command to the thermostat"""

        packet.sequence = self._get_sequence()

        self.logger.debug(
            "Queuing data, sequence=%d, action=%s, functional_domain=%s, attribute=%d",
            packet.sequence,
            str(packet.action),
            str(packet.functional_domain),
            packet.attribute,
        )

        await self.packet_queue.put(packet)

    def _empty_packet_queue(self):
        try:
            for _ in range(self.packet_queue.qsize()):
                self.packet_queue.get_nowait()
                self.packet_queue.task_done()
        except:  # pylint: disable=bare-except
            pass

    async def _queue_loop(self, loop_count=None):
        """Periodically send items from the queue"""
        while loop_count is None or loop_count > 0:
            if loop_count is not None:
                loop_count -= 1

            try:
                packet: Packet

                while packet := self.packet_queue.get_nowait():
                    if self.transport:
                        serialized_packet = packet.serialize()

                        self.logger.info("Sent data: %s", serialized_packet.hex(" "))

                        self.transport.write(serialized_packet)
            except asyncio.QueueEmpty:
                pass

            await asyncio.sleep(QUEUE_FREQUENCY)

    async def _update_status(self):
        await asyncio.sleep(2)

        await self.read_mac_address()
        await self.read_thermostat_status()
        await self.read_control()
        await self.read_sensors()
        await self.read_thermostat_name()
        await self.configure_cos()
        await self.read_dehumidification_setpoint()
        await self.read_humidification_setpoint()
        await self.sync()

    def connection_made(self, transport: asyncio.Transport):
        """Called when a connection has been made to the socket"""
        self.logger.info("Aprilaire connection made")

        self.transport = transport
        self._empty_packet_queue()

        asyncio.ensure_future(self._queue_loop())
        asyncio.ensure_future(self._update_status())

    def data_received(self, data: bytes) -> None:
        """Called when data has been received from the socket"""
        self.logger.info("Aprilaire data received %s", data.hex(" "))

        parsed_packets = Packet.parse(data)

        for packet in parsed_packets:
            self.logger.debug(
                "Received data, action=%s, functional_domain=%s, attribute=%d",
                str(packet.action),
                str(packet.functional_domain),
                packet.attribute,
            )

            if isinstance(packet, NackPacket):
                self.logger.error(
                    "Received NACK for attribute %d", packet.nack_attribute
                )
                continue

            if "error" in packet.data:
                error = packet.data["error"]

                if error != 0:
                    self.logger.error("Thermostat error: %d", error)

            if (
                packet.action == Action.COS
                and packet.functional_domain == FunctionalDomain.CONTROL
                and packet.attribute == 1
                and packet.data.get("mode") == 1
            ):
                self.logger.info("Re-reading control because of COS with mode==1")

                asyncio.ensure_future(self.read_control())

                continue

            if self.data_received_callback:
                asyncio.ensure_future(
                    self.data_received_callback(
                        packet.functional_domain, packet.attribute, packet.data
                    )
                )

    def connection_lost(self, exc: Exception | None) -> None:
        """Called when the connection to the socket has been lost"""
        self.logger.info("Aprilaire connection lost")

        if self.data_received_callback:
            asyncio.ensure_future(
                self.data_received_callback(
                    FunctionalDomain.NONE, 0, {"available": False}
                )
            )

        self.transport = None

        if self.reconnect_action:
            asyncio.ensure_future(self.reconnect_action())

    async def read_sensors(self):
        """Send a request for updated sensor data"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
        )

    async def read_control(self):
        """Send a request for updated control data"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
        )

    async def read_scheduling(self):
        """Send a request for updated scheduling data"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
        )

    async def update_mode(self, mode: int):
        """Send a request to update the mode"""
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": mode,
                    "fan_mode": 0,
                    "heat_setpoint": 0,
                    "cool_setpoint": 0,
                },
            )
        )

    async def update_fan_mode(self, fan_mode: int):
        """Send a request to update the fan mode"""
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 0,
                    "fan_mode": fan_mode,
                    "heat_setpoint": 0,
                    "cool_setpoint": 0,
                },
            )
        )

    async def update_setpoint(self, cool_setpoint: int, heat_setpoint: int):
        """Send a request to update the setpoint"""
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 0,
                    "fan_mode": 0,
                    "heat_setpoint": heat_setpoint,
                    "cool_setpoint": cool_setpoint,
                },
            )
        )

    async def set_hold(self, hold: int):
        """Send a request to set the hold status"""

        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.SCHEDULING,
                4,
                data={"hold": hold},
            )
        )

    async def set_dehumidification_setpoint(self, dehumidification_setpoint: int):
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                3,
                data={"dehumidification_setpoint": dehumidification_setpoint},
            )
        )

        await self.read_dehumidification_setpoint()

    async def set_humidification_setpoint(self, humidification_setpoint: int):
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                4,
                data={"humidification_setpoint": humidification_setpoint},
            )
        )

        await self.read_humidification_setpoint()

    async def set_fresh_air(self, mode: int, event: int):
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                5,
                data={"fresh_air_mode": mode, "fresh_air_event": event},
            )
        )

    async def set_air_cleaning(self, mode: int, event: int):
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                6,
                data={"air_cleaning_mode": mode, "air_cleaning_event": event},
            )
        )

    async def sync(self):
        """Send a request to sync data"""
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.STATUS,
                2,
                data={"synced": 1},
            )
        )

    async def configure_cos(self):
        """Send a request to configure the COS settings"""
        await self._send_packet(
            Packet(
                Action.WRITE,
                FunctionalDomain.STATUS,
                1,
                raw_data=[
                    1,  # Installer Thermostat Settings
                    0,  # Contractor Information
                    0,  # Air Cleaning Installer Variable
                    0,  # Humidity Control Installer Settings
                    0,  # Fresh Air Installer Settings
                    1,  # Thermostat Setpoint & Mode Settings
                    1,  # Dehumidification Setpoint
                    1,  # Humidification Setpoint
                    1,  # Fresh Air Settings
                    1,  # Air Cleaning Settings
                    1,  # Thermostat IAQ Available
                    0,  # Schedule Settings
                    1,  # Away Settings
                    0,  # Schedule Day
                    1,  # Schedule Hold
                    0,  # Heat Blast
                    0,  # Service Reminders Status
                    0,  # Alerts Status
                    0,  # Alerts Settings
                    0,  # Backlight Settings
                    1,  # Thermostat Location & Name
                    0,  # Reserved
                    1,  # Controlling Sensor Values
                    0,  # Over the air ODT update timeout
                    1,  # Thermostat Status
                    1,  # IAQ Status
                    1,  # Model & Revision
                    0,  # Support Module
                    0,  # Lockouts
                ],
            )
        )

    async def read_mac_address(self):
        """Send a request to get identification data (including MAC address)"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.IDENTIFICATION, 2)
        )

    async def read_thermostat_status(self):
        """Send a request for thermostat status"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 7)
        )

    async def read_thermostat_name(self):
        """Send a reques for the thermostat name"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.IDENTIFICATION, 5)
        )

    async def read_dehumidification_setpoint(self):
        """Send a request for the dehumidification setpoint"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 3)
        )

    async def read_humidification_setpoint(self):
        """Send a request for the humidification setpoint"""
        await self._send_packet(
            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 4)
        )


class AprilaireClient(SocketClient):
    """Client for sending/receiving data"""

    def __init__(
        self,
        host: str,
        port: int,
        data_received_callback: Callable[[dict[str, Any]], None],
        logger: Logger,
        reconnect_interval: int = None,
        retry_connection_interval: int = None,
    ) -> None:
        self.protocol: _AprilaireClientProtocol = None

        super().__init__(
            host,
            port,
            data_received_callback,
            logger,
            reconnect_interval,
            retry_connection_interval,
        )

        self.futures: dict[tuple[FunctionalDomain, int], list[asyncio.Future]] = {}

    async def _reconnect_with_delay(self):
        await super()._reconnect(self.retry_connection_interval)

    def create_protocol(self):
        return _AprilaireClientProtocol(
            self.data_received, self._reconnect_with_delay, self.logger
        )

    async def data_received(
        self, functional_domain: FunctionalDomain, attribute: int, data: dict[str, Any]
    ):
        """Called when data is received from the thermostat"""
        self.data_received_callback(data)

        if not functional_domain or not attribute:
            return

        future_key = (functional_domain, attribute)

        futures_to_complete = self.futures.pop(future_key, [])

        for future in futures_to_complete:
            try:
                future.set_result(data)
            except asyncio.exceptions.InvalidStateError:
                pass

    def state_changed(self):
        """Send data indicating the state as changed"""
        data = {
            "connected": self.connected,
            "stopped": self.stopped,
            "reconnecting": self.reconnecting,
        }

        self.data_received_callback(data)

    async def wait_for_response(
        self, functional_domain: FunctionalDomain, attribute: int, timeout: int = None
    ):
        """Wait for a response for a particular request"""

        loop = asyncio.get_event_loop()
        future = loop.create_future()

        future_key = (functional_domain, attribute)

        if future_key not in self.futures:
            self.futures[future_key] = []

        self.futures[future_key].append(future)

        try:
            return await asyncio.wait_for(future, timeout)
        except asyncio.exceptions.TimeoutError:
            self.logger.error(
                "Hit timeout of %d waiting for %s, %d",
                timeout,
                int(functional_domain),
                attribute,
            )
            return None

    async def read_sensors(self):
        """Send a request for updated sensor data"""
        await self.protocol.read_sensors()

    async def read_control(self):
        """Send a request for updated control data"""
        await self.protocol.read_control()

    async def read_scheduling(self):
        """Send a request for updated scheduling data"""
        await self.protocol.read_scheduling()

    async def update_mode(self, mode: int):
        """Send a request to update the mode"""
        await self.protocol.update_mode(mode)

    async def update_fan_mode(self, fan_mode: int):
        """Send a request to update the fan mode"""
        await self.protocol.update_fan_mode(fan_mode)

    async def update_setpoint(self, cool_setpoint: int, heat_setpoint: int):
        """Send a request to update the setpoint"""
        await self.protocol.update_setpoint(cool_setpoint, heat_setpoint)

    async def set_hold(self, hold: int):
        """Send a request to update the away status"""
        await self.protocol.set_hold(hold)

    async def sync(self):
        """Send a request to sync data"""
        await self.protocol.sync()

    async def read_mac_address(self):
        """Send a request to read the MAC address"""
        await self.protocol.read_mac_address()

    async def read_thermostat_status(self):
        """Send a request to read the thermostat status"""
        await self.protocol.read_thermostat_status()

    async def read_thermostat_name(self):
        """Send a request to read the thermostat name"""
        await self.protocol.read_thermostat_name()

    async def set_dehumidification_setpoint(self, dehumidification_setpoint: int):
        await self.protocol.set_dehumidification_setpoint(dehumidification_setpoint)

    async def set_humidification_setpoint(self, humidification_setpoint: int):
        await self.protocol.set_humidification_setpoint(humidification_setpoint)

    async def set_fresh_air(self, mode: int, event: int):
        await self.protocol.set_fresh_air(mode, event)

    async def set_air_cleaning(self, mode: int, event: int):
        await self.protocol.set_air_cleaning(mode, event)



"""Functions for handling response data from the thermostat"""

from __future__ import annotations

from collections.abc import Iterator
from crc import Calculator, Configuration
from enum import Enum
import math
from typing import Any

from pyaprilaire.const import Action, FunctionalDomain

crc_calculator = Calculator(
    Configuration(
        width=8,
        polynomial=0x31,
        init_value=0,
        final_xor_value=0,
        reverse_input=False,
        reverse_output=False,
    )
)


class ValueType(Enum):
    """Parsable value types from data"""

    INTEGER = 1
    INTEGER_REQUIRED = 2
    TEMPERATURE = 3
    TEMPERATURE_REQUIRED = 4
    HUMIDITY = 5
    MAC_ADDRESS = 6
    TEXT = 7


MAPPING = {
    Action.READ_RESPONSE: {
        FunctionalDomain.SETUP: {
            1: [
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                ("away_available", ValueType.INTEGER),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
            ]
        },
        FunctionalDomain.CONTROL: {
            1: [
                ("mode", ValueType.INTEGER_REQUIRED),
                ("fan_mode", ValueType.INTEGER_REQUIRED),
                ("heat_setpoint", ValueType.TEMPERATURE_REQUIRED),
                ("cool_setpoint", ValueType.TEMPERATURE_REQUIRED),
            ],
            3: [
                ("dehumidification_setpoint", ValueType.HUMIDITY),
            ],
            4: [
                ("humidification_setpoint", ValueType.HUMIDITY),
            ],
            5: [
                ("fresh_air_mode", ValueType.INTEGER),
                ("fresh_air_event", ValueType.INTEGER),
            ],
            6: [
                ("air_cleaning_mode", ValueType.INTEGER),
                ("air_cleaning_event", ValueType.INTEGER),
            ],
            7: [
                ("thermostat_modes", ValueType.INTEGER),
                ("air_cleaning_available", ValueType.INTEGER),
                ("ventilation_available", ValueType.INTEGER),
                ("dehumidification_available", ValueType.INTEGER),
                ("humidification_available", ValueType.INTEGER),
            ],
        },
        FunctionalDomain.SCHEDULING: {
            4: [
                ("hold", ValueType.INTEGER),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
            ],
        },
        FunctionalDomain.SENSORS: {
            1: [
                ("built_in_temperature_sensor_status", ValueType.INTEGER),
                ("built_in_temperature_sensor_value", ValueType.TEMPERATURE),
                (
                    "wired_remote_temperature_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "wired_remote_temperature_sensor_value",
                    ValueType.TEMPERATURE,
                ),
                (
                    "wired_outdoor_temperature_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "wired_outdoor_temperature_sensor_value",
                    ValueType.TEMPERATURE,
                ),
                ("built_in_humidity_sensor_status", ValueType.INTEGER),
                ("built_in_humidity_sensor_value", ValueType.HUMIDITY),
                ("rat_sensor_status", ValueType.INTEGER),
                ("rat_sensor_value", ValueType.TEMPERATURE),
                ("lat_sensor_status", ValueType.INTEGER),
                ("lat_sensor_value", ValueType.TEMPERATURE),
                (
                    "wireless_outdoor_temperature_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "wireless_outdoor_temperature_sensor_value",
                    ValueType.TEMPERATURE,
                ),
                (
                    "wireless_outdoor_humidity_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "wireless_outdoor_humidity_sensor_value",
                    ValueType.HUMIDITY,
                ),
            ],
            2: [
                (
                    "indoor_temperature_controlling_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "indoor_temperature_controlling_sensor_value",
                    ValueType.TEMPERATURE,
                ),
                (
                    "outdoor_temperature_controlling_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "outdoor_temperature_controlling_sensor_value",
                    ValueType.TEMPERATURE,
                ),
                (
                    "indoor_humidity_controlling_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "indoor_humidity_controlling_sensor_value",
                    ValueType.HUMIDITY,
                ),
                (
                    "outdoor_humidity_controlling_sensor_status",
                    ValueType.INTEGER,
                ),
                (
                    "outdoor_humidity_controlling_sensor_value",
                    ValueType.HUMIDITY,
                ),
            ],
        },
        FunctionalDomain.STATUS: {
            2: [
                ("synced", ValueType.INTEGER),
            ],
            6: [
                ("heating_equipment_status", ValueType.INTEGER),
                ("cooling_equipment_status", ValueType.INTEGER),
                ("progressive_recovery", ValueType.INTEGER),
                ("fan_status", ValueType.INTEGER),
            ],
            7: [
                ("dehumidification_status", ValueType.INTEGER),
                ("humidification_status", ValueType.INTEGER),
                ("ventilation_status", ValueType.INTEGER),
                ("air_cleaning_status", ValueType.INTEGER),
            ],
            8: [
                ("error", ValueType.INTEGER),
            ],
        },
        FunctionalDomain.IDENTIFICATION: {
            1: [
                ("hardware_revision", ValueType.INTEGER),
                ("firmware_major_revision", ValueType.INTEGER),
                ("firmware_minor_revision", ValueType.INTEGER),
                ("protocol_major_revision", ValueType.INTEGER),
                ("model_number", ValueType.INTEGER),
                ("gainspan_firmware_major_revision", ValueType.INTEGER),
                ("gainspan_firmware_minor_revision", ValueType.INTEGER),
            ],
            2: [
                ("mac_address", ValueType.MAC_ADDRESS),
            ],
            4: [
                ("location", ValueType.TEXT, 7),
                ("name", ValueType.TEXT, 15),
            ],
            5: [
                ("location", ValueType.TEXT, 7),
                ("name", ValueType.TEXT, 15),
            ],
        },
    }
}

MAPPING[Action.COS] = MAPPING[Action.READ_RESPONSE]
MAPPING[Action.WRITE] = MAPPING[Action.READ_RESPONSE]
MAPPING[Action.READ_REQUEST] = MAPPING[Action.READ_RESPONSE]


class Packet:
    def __init__(
        self,
        action: Action,
        functional_domain: FunctionalDomain,
        attribute: int,
        revision: int = 1,
        sequence: int = 0,
        count: int = 0,
        data: dict[str, Any] = None,
        raw_data: list[int] = None,
    ):
        self.action = action
        self.functional_domain = functional_domain
        self.attribute = attribute
        self.revision = revision
        self.sequence = sequence
        self.count = count
        self.data = data or {}
        self.raw_data = raw_data

    @classmethod
    def parse(self, data: bytes) -> Iterator[Packet]:
        data_index = 0

        while data_index < len(data):
            revision = data[data_index]
            sequence = data[data_index + 1]
            count = data[data_index + 2] << 2 | data[data_index + 3]

            action = int(data[data_index + 4])
            functional_domain = int(data[data_index + 5])
            attribute = int(data[data_index + 6])

            try:
                action = Action(action)
                functional_domain = FunctionalDomain(functional_domain)
            except:
                data_index += count + 5
                continue

            if action == Action.NACK:
                nack_attribute = int(data[data_index + 5])

                yield NackPacket(nack_attribute)

                data_index += count + 5
                continue

            if (
                action not in MAPPING
                or functional_domain not in MAPPING[action]
                or attribute not in MAPPING[action][functional_domain]
            ):
                data_index += count + 5
                continue

            packet = Packet(
                action, functional_domain, attribute, revision, sequence, count
            )

            # Skip header
            final_index = data_index + count + 3
            payload_start_index = data_index
            data_index += 7
            attribute_index = 0

            while data_index <= final_index:
                if attribute_index >= len(
                    MAPPING[action][functional_domain][attribute]
                ):
                    data_index += 1
                    pass
                else:
                    attribute_info = MAPPING[action][functional_domain][attribute][
                        attribute_index
                    ]

                    (attribute_name, value_type, extra_attribute_info) = (
                        attribute_info[0],
                        attribute_info[1],
                        attribute_info[2:],
                    )

                    if attribute_name is None or value_type is None:
                        data_index += 1
                        attribute_index += 1
                        continue

                    data_value = data[data_index]

                    if value_type == ValueType.INTEGER:
                        packet.data[attribute_name] = data_value
                        data_index += 1
                    elif value_type == ValueType.INTEGER_REQUIRED:
                        if data_value is not None and data_value != 0:
                            packet.data[attribute_name] = data_value
                        data_index += 1
                    elif value_type == ValueType.HUMIDITY:
                        packet.data[attribute_name] = self._decode_humidity(data_value)
                        data_index += 1
                    elif value_type == ValueType.TEMPERATURE:
                        packet.data[attribute_name] = self._decode_temperature(
                            data_value
                        )
                        data_index += 1
                    elif value_type == ValueType.TEMPERATURE_REQUIRED:
                        if data_value is not None and data_value != 0:
                            packet.data[attribute_name] = self._decode_temperature(
                                data_value
                            )
                        data_index += 1
                    elif value_type == ValueType.MAC_ADDRESS:
                        mac_address_components = []

                        for _ in range(0, 6):
                            mac_address_components.append(f"{data[data_index]:x}")
                            data_index += 1

                        packet.data[attribute_name] = ":".join(mac_address_components)
                    elif value_type == ValueType.TEXT:
                        text_length = extra_attribute_info[0]

                        text = ""

                        for _ in range(0, text_length):
                            current_value = (
                                " " if data[data_index] == 0 else chr(data[data_index])
                            )
                            text += current_value
                            data_index += 1

                        data_index += 1

                        text = text.strip(" ")

                        packet.data[attribute_name] = text

                    attribute_index += 1

            crc = data[data_index]

            if Packet._verify_crc(data[payload_start_index:data_index], crc):
                yield packet

            data_index += 1

    @classmethod
    def _generate_crc(self, lst: list[int]):
        """Generate a CRC checksum"""
        return crc_calculator.checksum(bytes(lst))

    @classmethod
    def _verify_crc(self, lst: list[int], crc: int):
        """Verify a CRC checksum"""
        return crc_calculator.verify(bytes(lst), crc)

    @classmethod
    def _encode_temperature(self, temperature: float) -> int:
        """Encode a temperature value for sending to the thermostat"""
        is_negative = temperature < 0
        is_fraction = temperature % 1 >= 0.5

        return (
            math.floor(abs(temperature))
            + (64 if is_fraction else 0)
            + (128 if is_negative else 0)
        )

    @classmethod
    def _decode_temperature(self, raw_value: int) -> float:
        """Decode a temperature value from the thermostat"""
        temperature_value = float(int(raw_value & 63))

        raw_value = raw_value >> 6
        has_fraction = bool(raw_value & 1)
        if has_fraction:
            temperature_value += 0.5

        raw_value = raw_value >> 1
        is_positive = raw_value & 1 == 0
        if not is_positive:
            temperature_value = -temperature_value

        return temperature_value

    @classmethod
    def _decode_humidity(self, raw_value: int) -> int:
        """Decode a humidity value from the thermostat"""
        if raw_value <= 0 or raw_value >= 100:
            return None
        return raw_value

    @classmethod
    def _encode_int_value(self, value: int):
        return ((value >> 8) & 0xFF, value & 0xFF)

    def serialize(self) -> bytes:
        if isinstance(self, NackPacket):
            payload = [int(Action.NACK), self.nack_attribute]
        else:
            payload = [int(self.action), int(self.functional_domain), self.attribute]

            if self.raw_data is not None:
                payload.extend(self.raw_data)
            elif (
                self.action == Action.WRITE
                or self.action == Action.READ_RESPONSE
                or self.action == Action.COS
            ):
                for attribute_info in MAPPING[self.action][self.functional_domain][
                    self.attribute
                ]:
                    (attribute_name, value_type, extra_attribute_info) = (
                        attribute_info[0],
                        attribute_info[1],
                        attribute_info[2:],
                    )

                    data_value = self.data.get(attribute_name)

                    if (
                        value_type == ValueType.INTEGER
                        or value_type == ValueType.INTEGER_REQUIRED
                        or value_type == ValueType.HUMIDITY
                    ):
                        payload.append(data_value)
                    elif (
                        value_type == ValueType.TEMPERATURE
                        or value_type == ValueType.TEMPERATURE_REQUIRED
                    ):
                        payload.append(self._encode_temperature(data_value))
                    elif value_type == ValueType.MAC_ADDRESS:
                        payload.extend(data_value)
                    elif value_type == ValueType.TEXT:
                        text_length = extra_attribute_info[0]

                        for i in range(0, text_length + 1):
                            if i >= len(data_value):
                                payload.append(0)
                            else:
                                payload.append(ord(data_value[i]))
                    else:
                        payload.append(0)

        (payload_length_high, payload_length_low) = self._encode_int_value(len(payload))
        result = [1, self.sequence, payload_length_high, payload_length_low]
        result.extend(payload)
        result.append(self._generate_crc(result))
        return bytes(result)

    def __eq__(self, other):
        return (
            self.action == other.action
            and self.functional_domain == other.functional_domain
            and self.attribute == other.attribute
            and self.data == other.data
        )


class NackPacket(Packet):
    def __init__(
        self,
        nack_attribute: int,
        revision: int = 1,
        sequence: int = 0,
        count: int = 0,
    ):
        super().__init__(
            Action.NACK, FunctionalDomain.NACK, 0, revision, sequence, count
        )

        self.nack_attribute = nack_attribute



from pyaprilaire.client import _AprilaireClientProtocol, AprilaireClient
from pyaprilaire.const import Action, FunctionalDomain
from pyaprilaire.packet import Packet

import asyncio
import logging

import unittest
from unittest.mock import patch, AsyncMock, Mock


class Test_Protocol(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.logger = logging.getLogger()
        self.logger.propagate = False

        self.data_received_mock = AsyncMock()
        self.reconnect_action_mock = AsyncMock()

        self.protocol = _AprilaireClientProtocol(
            self.data_received_mock, self.reconnect_action_mock, self.logger
        )

    def test_connection_made(self):
        self.protocol._queue_loop = AsyncMock()
        self.protocol._update_status = AsyncMock()

        self.protocol.connection_made(None)

        self.assertEqual(self.protocol._queue_loop.call_count, 1)
        self.assertEqual(self.protocol._update_status.call_count, 1)

    async def test_update_status(self):
        sleep_mock = AsyncMock()

        with patch("asyncio.sleep", new=sleep_mock):
            await self.protocol._update_status()

        self.assertEqual(self.protocol.packet_queue.qsize(), 9)
        self.assertEqual(sleep_mock.call_count, 1)

    async def test_queue_loop(self):
        await self.protocol.read_control()
        await self.protocol.read_scheduling()

        sleep_mock = AsyncMock()
        self.protocol.transport = Mock(asyncio.Transport)

        with patch("asyncio.sleep", new=sleep_mock):
            await self.protocol._queue_loop(loop_count=1)

        self.assertEqual(sleep_mock.call_count, 1)
        self.assertEqual(self.protocol.transport.write.call_count, 2)

    def test_data_received(self):
        self.protocol.data_received(bytes([1, 1, 0, 7, 3, 2, 1, 1, 2, 10, 20, 107]))

        self.assertEqual(self.data_received_mock.call_count, 1)

        (functional_domain, attribute, data) = self.data_received_mock.call_args[0]

        self.assertEqual(functional_domain, FunctionalDomain.CONTROL)
        self.assertEqual(attribute, 1)
        self.assertDictEqual(
            data,
            {
                "mode": 1,
                "fan_mode": 2,
                "heat_setpoint": 10,
                "cool_setpoint": 20,
            },
        )

    def test_data_received_nack(self):
        with self.assertLogs(self.logger, level="ERROR") as cm:
            self.protocol.data_received(bytes([1, 1, 0, 2, 6, 1, 0]))

        self.assertEqual(cm.output, ["ERROR:root:Received NACK for attribute 1"])

        self.assertEqual(self.data_received_mock.call_count, 0)

    def test_data_received_error(self):
        with self.assertLogs(self.logger, level="ERROR") as cm:
            self.protocol.data_received(bytes([1, 1, 0, 4, 3, 7, 8, 2, 149]))

        self.assertEqual(cm.output, ["ERROR:root:Thermostat error: 2"])

        self.assertEqual(self.data_received_mock.call_count, 1)

        (functional_domain, attribute, data) = self.data_received_mock.call_args[0]

        self.assertEqual(functional_domain, FunctionalDomain.STATUS)
        self.assertEqual(attribute, 8)
        self.assertDictEqual(
            data,
            {
                "error": 2,
            },
        )

    def test_mode_re_read(self):
        self.protocol.read_control = AsyncMock()

        self.protocol.data_received(bytes([1, 1, 0, 7, 5, 2, 1, 1, 2, 10, 20, 127]))

        self.assertEqual(self.protocol.read_control.call_count, 1)

    def test_connection_lost(self):
        self.protocol.connection_lost(None)

        self.assertEqual(self.data_received_mock.call_count, 1)

        (functional_domain, attribute, data) = self.data_received_mock.call_args[0]

        self.assertEqual(functional_domain, FunctionalDomain.NONE)
        self.assertEqual(attribute, 0)
        self.assertDictEqual(
            data,
            {
                "available": False,
            },
        )

    def test_get_sequence(self):
        sequence = self.protocol._get_sequence()
        self.assertEqual(sequence, 1)

        sequence = self.protocol._get_sequence()
        self.assertEqual(sequence, 2)

        self.protocol.sequence = 127
        sequence = self.protocol._get_sequence()
        self.assertEqual(sequence, 0)

        sequence = self.protocol._get_sequence()
        self.assertEqual(sequence, 1)

    async def test_send_packet(self):
        self.protocol.packet_queue.put = AsyncMock()
        self.protocol.packet_queue.put_nowait = Mock()

        original_packet = Packet(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            data={
                "mode": 1,
                "fan_mode": 0,
                "heat_setpoint": 0,
                "cool_setpoint": 0,
            },
        )

        await self.protocol._send_packet(original_packet)

        self.assertEqual(self.protocol.packet_queue.put.call_count, 1)

        (sent_packet) = self.protocol.packet_queue.put.call_args[0][0]

        self.assertEqual(original_packet, sent_packet)

    def assertPacketQueueContains(self, packet: Packet):
        queue_items = list(self.protocol.packet_queue._queue)

        self.assertEqual(any(qp == packet for qp in queue_items), True)

    async def test_read_sensors(self):
        await self.protocol.read_sensors()

        self.assertPacketQueueContains(
            Packet(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
        )

    async def test_read_control(self):
        await self.protocol.read_control()

        self.assertPacketQueueContains(
            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
        )

    async def test_read_scheduling(self):
        await self.protocol.read_scheduling()

        self.assertPacketQueueContains(
            Packet(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
        )

    async def test_update_mode(self):
        await self.protocol.update_mode(1)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 1,
                    "fan_mode": 0,
                    "heat_setpoint": 0,
                    "cool_setpoint": 0,
                },
            )
        )

    async def test_update_fan_mode(self):
        await self.protocol.update_fan_mode(1)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 0,
                    "fan_mode": 1,
                    "heat_setpoint": 0,
                    "cool_setpoint": 0,
                },
            )
        )

    async def test_update_setpoint(self):
        await self.protocol.update_setpoint(10, 20)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 0,
                    "fan_mode": 0,
                    "heat_setpoint": 20,
                    "cool_setpoint": 10,
                },
            )
        )

    async def test_set_hold(self):
        await self.protocol.set_hold(1)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.SCHEDULING,
                4,
                data={
                    "hold": 1,
                },
            )
        )

    async def test_sync(self):
        await self.protocol.sync()

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.STATUS,
                2,
                data={
                    "synced": 1,
                },
            )
        )

    async def test_configure_cos(self):
        pass

    async def test_read_mac_address(self):
        await self.protocol.read_mac_address()

        self.assertPacketQueueContains(
            Packet(
                Action.READ_REQUEST,
                FunctionalDomain.IDENTIFICATION,
                2,
            )
        )

    async def test_read_thermostat_status(self):
        await self.protocol.read_thermostat_status()

        self.assertPacketQueueContains(
            Packet(
                Action.READ_REQUEST,
                FunctionalDomain.CONTROL,
                7,
            )
        )

    async def test_read_thermostat_name(self):
        await self.protocol.read_thermostat_name()

        self.assertPacketQueueContains(
            Packet(
                Action.READ_REQUEST,
                FunctionalDomain.IDENTIFICATION,
                5,
            )
        )

    def test_empty_packet_queue(self):
        self.protocol.packet_queue.put_nowait({})
        self.protocol.packet_queue.put_nowait({})

        self.assertEqual(self.protocol.packet_queue.qsize(), 2)

        self.protocol._empty_packet_queue()

        self.assertEqual(self.protocol.packet_queue.qsize(), 0)

    def test_empty_packet_queue_error(self):
        self.protocol.packet_queue.get_nowait = Mock(side_effect=Exception)

        self.protocol.packet_queue.put_nowait({})
        self.protocol.packet_queue.put_nowait({})

        self.assertEqual(self.protocol.packet_queue.qsize(), 2)

        self.protocol._empty_packet_queue()

        self.assertEqual(self.protocol.packet_queue.qsize(), 2)


class Test_Client(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.logger = logging.getLogger()
        self.logger.propagate = False

        self.data_received_mock = Mock()

        self.client = AprilaireClient(
            None, None, self.data_received_mock, self.logger, 10, 10
        )
        self.client.protocol = _AprilaireClientProtocol(Mock(), Mock(), self.logger)

    def test_create_protocol(self):
        protocol = self.client.create_protocol()

        self.assertIsInstance(protocol, _AprilaireClientProtocol)

    async def test_data_received(self):
        functional_domain = FunctionalDomain.CONTROL
        attribute = 1
        data = {"testKey": "testValue"}

        loop = asyncio.get_event_loop()
        future = loop.create_future()

        future_key = (functional_domain, attribute)

        if future_key not in self.client.futures:
            self.client.futures[future_key] = []

        self.client.futures[future_key].append(future)

        await self.client.data_received(functional_domain, attribute, data)

        self.assertEqual(self.data_received_mock.call_count, 1)
        self.assertEqual(self.data_received_mock.call_args[0][0], data)

        self.assertEqual(future.result(), data)

    async def test_data_received_empty(self):
        await self.client.data_received(None, None, None)

    async def test_data_received_state_error(self):
        functional_domain = FunctionalDomain.CONTROL
        attribute = 1

        loop = asyncio.get_event_loop()
        future = loop.create_future()

        future_key = (functional_domain, attribute)

        if future_key not in self.client.futures:
            self.client.futures[future_key] = []

        self.client.futures[future_key].append(future)

        future.set_result({})

        await self.client.data_received(functional_domain, attribute, {})

    def test_state_changed(self):
        self.client.connected = True
        self.client.stopped = True
        self.client.reconnecting = True

        self.client.state_changed()

        self.assertEqual(self.data_received_mock.call_count, 1)
        self.assertEqual(
            self.data_received_mock.call_args[0][0],
            {"connected": True, "stopped": True, "reconnecting": True},
        )

    def assertPacketQueueContains(self, packet: Packet):
        queue_items = list(self.client.protocol.packet_queue._queue)

        self.assertEqual(any(qp == packet for qp in queue_items), True)

    async def test_read_sensors(self):
        await self.client.read_sensors()

        self.assertPacketQueueContains(
            Packet(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
        )

    async def test_read_control(self):
        await self.client.read_control()

        self.assertPacketQueueContains(
            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
        )

    async def test_read_scheduling(self):
        await self.client.read_scheduling()

        self.assertPacketQueueContains(
            Packet(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
        )

    async def test_update_mode(self):
        await self.client.update_mode(1)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 1,
                    "fan_mode": 0,
                    "heat_setpoint": 0,
                    "cool_setpoint": 0,
                },
            )
        )

    async def test_update_fan_mode(self):
        await self.client.update_fan_mode(1)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 0,
                    "fan_mode": 1,
                    "heat_setpoint": 0,
                    "cool_setpoint": 0,
                },
            )
        )

    async def test_update_setpoint(self):
        await self.client.update_setpoint(10, 20)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                1,
                data={
                    "mode": 0,
                    "fan_mode": 0,
                    "heat_setpoint": 20,
                    "cool_setpoint": 10,
                },
            )
        )

    async def test_set_hold(self):
        await self.client.set_hold(1)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.SCHEDULING,
                4,
                data={
                    "hold": 1,
                },
            )
        )

    async def test_set_dehumidification_setpoint(self):
        await self.client.set_dehumidification_setpoint(50)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                3,
                data={
                    "dehumidification_setpoint": 50,
                },
            )
        )

    async def test_set_humidification_setpoint(self):
        await self.client.set_humidification_setpoint(50)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                4,
                data={
                    "humidification_setpoint": 50,
                },
            )
        )

    async def test_set_fresh_air(self):
        await self.client.set_fresh_air(1, 3)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                5,
                data={
                    "fresh_air_mode": 1,
                    "fresh_air_event": 3,
                },
            )
        )

    async def test_set_air_cleaning(self):
        await self.client.set_air_cleaning(1, 3)

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.CONTROL,
                6,
                data={
                    "air_cleaning_mode": 1,
                    "air_cleaning_event": 3,
                },
            )
        )

    async def test_sync(self):
        await self.client.sync()

        self.assertPacketQueueContains(
            Packet(
                Action.WRITE,
                FunctionalDomain.STATUS,
                2,
                data={
                    "synced": 1,
                },
            )
        )

    async def test_configure_cos(self):
        pass

    async def test_read_mac_address(self):
        await self.client.read_mac_address()

        self.assertPacketQueueContains(
            Packet(
                Action.READ_REQUEST,
                FunctionalDomain.IDENTIFICATION,
                2,
            )
        )

    async def test_read_thermostat_status(self):
        await self.client.read_thermostat_status()

        self.assertPacketQueueContains(
            Packet(
                Action.READ_REQUEST,
                FunctionalDomain.CONTROL,
                7,
            )
        )

    async def test_read_thermostat_name(self):
        await self.client.read_thermostat_name()

        self.assertPacketQueueContains(
            Packet(
                Action.READ_REQUEST,
                FunctionalDomain.IDENTIFICATION,
                5,
            )
        )

    async def test_wait_for_response_success(self):
        wait_for_mock = AsyncMock(return_value=True)

        with patch("asyncio.wait_for", new=wait_for_mock):
            wait_for_response_result = await self.client.wait_for_response(
                FunctionalDomain.CONTROL, 1, 1
            )

        self.assertEqual(wait_for_response_result, True)

    async def test_wait_for_response_timeout(self):
        wait_for_mock = AsyncMock(side_effect=asyncio.exceptions.TimeoutError)

        with (
            patch("asyncio.wait_for", new=wait_for_mock),
            self.assertLogs(self.logger, level="ERROR") as cm,
        ):
            wait_for_response_result = await self.client.wait_for_response(
                FunctionalDomain.CONTROL, 1, 1
            )

        self.assertEqual(cm.output, ["ERROR:root:Hit timeout of 1 waiting for 2, 1"])

        self.assertEqual(wait_for_response_result, None)

    async def test_reconnect_with_delay(self):
        reconnect_mock = AsyncMock()

        with patch(
            "pyaprilaire.socket_client.SocketClient._reconnect", new=reconnect_mock
        ):
            await self.client._reconnect_with_delay()

            self.assertEqual(reconnect_mock.call_count, 1)
            self.assertEqual(
                reconnect_mock.call_args[0][0],
                self.client.retry_connection_interval,
            )



"""Client for connecting to the Aprilaire thermostat socket"""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import Any
from logging import Logger


class SocketClient:
    """Client for connecting to the Aprilaire thermostat socket"""

    def __init__(
        self,
        host: str,
        port: int,
        data_received_callback: Callable[[dict[str, Any]], None],
        logger: Logger,
        reconnect_interval: int = None,
        retry_connection_interval: int = None,
    ) -> None:
        """Initialize client"""
        self.host = host
        self.port = port
        self.data_received_callback = data_received_callback
        self.logger = logger
        self.data: dict[str, Any] = {}
        self.reconnect_interval = reconnect_interval
        self.retry_connection_interval = retry_connection_interval

        self.connected = False
        self.stopped = True
        self.reconnecting = False
        self.cancelled = False
        self.reconnect_break_future: asyncio.Future = None

        self.protocol: asyncio.Protocol = None

    async def _reconnect_loop(self):
        """Wait for cancellable reconnect interval to pass, and perform reconnect"""
        if not self.reconnect_interval:
            return

        while True:
            if self.stopped or not self.connected:
                break

            if not self.reconnect_break_future:
                loop = asyncio.get_event_loop()
                self.reconnect_break_future = loop.create_future()

            try:
                await asyncio.wait_for(
                    self.reconnect_break_future, self.reconnect_interval
                )
                break
            except asyncio.exceptions.CancelledError:
                break
            except asyncio.exceptions.TimeoutError:
                await self._reconnect(10)

    def _cancel_reconnect_loop(self):
        """Cancel the loop which does periodic reconnection"""
        if self.reconnect_break_future:
            try:
                self.reconnect_break_future.set_result(True)
            except asyncio.exceptions.InvalidStateError:
                pass
            self.reconnect_break_future = None

    def _disconnect(self):
        """Disconnect from the socket"""
        self._cancel_reconnect_loop()

        self.connected = False

        self.state_changed()

        if self.protocol and self.protocol.transport:
            self.protocol.transport.close()

    async def _reconnect(self, connect_wait_period: int = 0):
        """Reconnect to the socket"""

        if self.reconnecting:
            return

        self.reconnecting = True

        self.state_changed()

        # Ensure already disconnected
        self._disconnect()

        if connect_wait_period is not None and connect_wait_period > 0:
            await asyncio.sleep(connect_wait_period)

        self.protocol = self.create_protocol()

        while True:
            if self.stopped or self.cancelled:
                break

            try:
                await asyncio.get_event_loop().create_connection(
                    lambda: self.protocol,
                    self.host,
                    self.port,
                )

                self.connected = True
                self.reconnecting = False

                self.state_changed()

                asyncio.ensure_future(self._reconnect_loop())

                break

            except Exception as exc:  # pylint: disable=broad-except
                self.logger.error("Failed to connect to thermostat: %s", str(exc))

                if not self.stopped:
                    await asyncio.sleep(self.retry_connection_interval)

    async def start_listen(self):
        """Start listening to the socket"""

        self.stopped = False

        self.state_changed()

        await self._reconnect()

    def stop_listen(self):
        """Stop listening to the socket"""

        self.stopped = True

        self.state_changed()

        self._disconnect()

    def create_protocol(self) -> asyncio.Protocol:
        """Create the socket protocol (implemented in derived class)"""
        return None

    def state_changed(self):
        """Handle a state change (implemented in derived class)"""
        pass



from pyaprilaire.client import SocketClient

import asyncio
from asyncio import Protocol

import logging

import unittest
from unittest.mock import patch, AsyncMock, Mock


class Test_Socket_Client(unittest.IsolatedAsyncioTestCase):
    def patch_socket(func):
        async def wrapper(*args, **kwargs):
            state_changed_mock = Mock()
            create_connection_mock = AsyncMock()
            create_protocol_mock = Mock(spec=Protocol)

            with (
                patch(
                    "asyncio.BaseEventLoop.create_connection",
                    new=create_connection_mock,
                ),
                patch(
                    "pyaprilaire.socket_client.SocketClient.state_changed",
                    new=state_changed_mock,
                ),
                patch(
                    "pyaprilaire.socket_client.SocketClient.create_protocol",
                    new=create_protocol_mock,
                ),
            ):
                await func(*args, **kwargs)

        return wrapper

    def setUp(self):
        self.logger = logging.getLogger()
        self.logger.propagate = False
        self.client = SocketClient(None, None, None, self.logger)

    @patch_socket
    async def test_client_status(self):
        await self.client.start_listen()

        self.assertFalse(self.client.stopped)
        self.assertTrue(self.client.connected)
        self.assertFalse(self.client.reconnecting)

        self.client._disconnect()

        self.assertFalse(self.client.stopped)
        self.assertFalse(self.client.connected)
        self.assertFalse(self.client.reconnecting)

        self.client.stop_listen()

        self.assertTrue(self.client.stopped)
        self.assertFalse(self.client.connected)
        self.assertFalse(self.client.reconnecting)

    @patch_socket
    async def test_reconnect_loop(self):
        self.client.reconnect_interval = 0.01

        async def _reconnect_nowait(self: SocketClient, connect_wait_period: int = 0):
            self.connected = True
            self.reconnecting = False

        with patch(
            "pyaprilaire.socket_client.SocketClient._reconnect", new=_reconnect_nowait
        ):
            await self.client.start_listen()
            await self.client._reconnect_loop()

        self.assertFalse(self.client.stopped)
        self.assertTrue(self.client.connected)
        self.assertFalse(self.client.reconnecting)

    @patch_socket
    async def test_reconnect_loop_cancelled(self):
        self.client.reconnect_interval = 0.01

        async def _reconnect_nowait(self: SocketClient, connect_wait_period: int = 0):
            self.connected = True
            self.reconnecting = False

        wait_for_mock = AsyncMock(side_effect=asyncio.exceptions.CancelledError)

        with patch(
            "pyaprilaire.socket_client.SocketClient._reconnect", new=_reconnect_nowait
        ), patch("asyncio.wait_for", new=wait_for_mock):
            await self.client.start_listen()
            await self.client._reconnect_loop()

        self.assertFalse(self.client.stopped)
        self.assertTrue(self.client.connected)
        self.assertFalse(self.client.reconnecting)

    @patch_socket
    async def test_reconnect_loop_stopped(self):
        self.client.reconnect_interval = 0.01
        self.client.connected = False
        self.client.stopped = True

        await self.client._reconnect_loop()

        self.assertTrue(self.client.stopped)
        self.assertFalse(self.client.connected)
        self.assertFalse(self.client.reconnecting)

    @patch_socket
    async def test_cancel_reconnect_loop(self):
        self.client.reconnect_interval = 1
        self.client.connected = True
        self.client.stopped = False

        async def cancel_reconnect_loop():
            while not self.client.reconnect_break_future:
                await asyncio.sleep(0.01)
            self.client._cancel_reconnect_loop()

        await asyncio.gather(cancel_reconnect_loop(), self.client._reconnect_loop())

    @patch_socket
    async def test_cancel_reconnect_loop_state_error(self):
        loop = asyncio.get_event_loop()
        future = loop.create_future()

        self.client.reconnect_break_future = future

        future.set_result(None)

        self.client._cancel_reconnect_loop()

    @patch_socket
    async def test_reconnect(self):
        self.client.stopped = False

        sleep_mock = AsyncMock()

        with patch("asyncio.sleep", new=sleep_mock):
            await self.client._reconnect(10)

        self.assertEqual(sleep_mock.await_count, 1)
        self.assertFalse(self.client.stopped)
        self.assertTrue(self.client.connected)
        self.assertFalse(self.client.reconnecting)

    async def test_reconnect_exception(self):
        self.client.stopped = False

        sleep_mock = AsyncMock()

        def create_connection(*args, **kwargs):
            self.client.cancelled = True
            raise Exception("Test failure")

        create_connection_mock = AsyncMock(side_effect=create_connection)

        with (
            patch("asyncio.sleep", new=sleep_mock),
            patch(
                "asyncio.BaseEventLoop.create_connection", new=create_connection_mock
            ),
            self.assertLogs(self.logger, level="ERROR") as cm,
        ):
            await self.client._reconnect(10)

        self.assertEqual(
            cm.output, ["ERROR:root:Failed to connect to thermostat: Test failure"]
        )

        self.assertEqual(sleep_mock.await_count, 2)
        self.assertFalse(self.client.stopped)
        self.assertFalse(self.client.connected)
        self.assertTrue(self.client.reconnecting)

    @patch_socket
    async def test_reconnect_reconnecting(self):
        self.client.stopped = False
        self.client.connected = True
        self.client.reconnecting = True

        sleep_mock = AsyncMock()

        with patch("asyncio.sleep", new=sleep_mock):
            await self.client._reconnect(10)

        self.assertEqual(sleep_mock.await_count, 0)
        self.assertFalse(self.client.stopped)
        self.assertTrue(self.client.connected)
        self.assertTrue(self.client.reconnecting)
